/* AVERAGING.PRG 

These Gauss procedures compute the Mallows Model Average (MMA) autoregressive least-squares estimates, 
as described in the paper "Averaging Estimators for Autoregressions with a Near Unit Root"

written by
Bruce E. Hansen
Department of Economics
Univerity of Wisconsin

There are two procedures.  

The first procedure takes the AR order as given, and computes the weighted average of the models
with and without the unit root imposed.

{beta,beta0,beta1,w}=URaveraging(y,k,p);

Inputs:
y			nx1 time series
k			autoregressive order (non-negative integer)
p			trend order (non-negative integer)

Outputs:
beta			averaging parameter estimate
beta0			restricted parameter estimate (unit root imposed)
beta1			unrestricted parameter estimate
w			weight for unrestricted estimate vector

The dependent variable is Dy(t)

The order of the parameter are:

Constant
Trend
y(t-1)
Dy(t-1)
...
Dy(t-k)

****************************************************************************************

The second procedure computes the general averaging estimator, over all autoregressive orders and over the unit root restriction.

{beta,w}=Gaveraging(y,K,p);

Inputs:
y			nx1 time series
K			maxmimum autoregressive order (non-negative integer)
p			trend order (non-negative integer)

Outputs:
beta			averaging parameter estimate
w			weight vector

The ordering of the parameters is the same.

The weight vector has (2K+2) elements. 
The first K+1 are the weights for the unrestricted AR(j) models, j=0,...,K. 
The second K+1 are the weights for the restricted (unit root) AR(j) models, j=0,...,K. 


*/


proc (4) = URaveraging(y,k,p);
local n,dy,x,xxi,beta1,r,beta0,e,fc,beta,w;
  n=rows(y)-k-1;
  dy=y[k+2:n+k+1]-y[k+1:n+k];
  x=(seqa(1,1,n).^(seqa(0,1,p+1)'))~y[k+1:n+k];
  if k>0;
    for j (1,k,1);
      x=x~(y[k+2-j:n+k+1-j]-y[k+1-j:n+k-j]);
    endfor;
  endif;
  xxi=invpd(moment(x,0));
  beta1=xxi*(x'dy);
  r=zeros(k+2+p,2);
  r[p+1,1]=1;
  r[p+2,2]=2;
  beta0=beta1-xxi*r*invpd(r'xxi*r)*(r'beta1);
  beta0[p+1:p+2]=0|0;
  e=dy-x*beta1; 
  fc=(beta1[p+1 p+2]'invpd(xxi[p+1 p+2,p+1 p+2])*beta1[p+1 p+2])*n/(e'e);
  if fc>2; w=1-2/fc; else; w=0; endif;
  beta=beta1.*w+beta0.*(1-w);
retp(beta,beta0,beta1,w);
endp;

y=cumsumc(rndn(100,1));
k=4;
p=1;
{beta,beta0,beta1,w}=URaveraging(y,k,p);
"";
{betaw,w}=Gaveraging(y,k,p);
w;
beta~betaw;


proc (2) = Gaveraging(y,k,p);
local n,dy,x,xxi,beta,e0,e1,beta0,beta1,r0,r00,r1,b1,b0,aa,sig,r,alpha0,kk,w,u1,u2,u3,u4,ret;
  n=rows(y)-k-1;
  dy=y[k+2:n+k+1]-y[k+1:n+k];
  x=(seqa(1,1,n).^(seqa(0,1,p+1)'))~y[k+1:n+k];
  if k>0;
    for j (1,k,1);
      x=x~(y[k+2-j:n+k+1-j]-y[k+1-j:n+k-j]);
    endfor;
  endif;
  xxi=invpd(moment(x,0));
  beta=xxi*(x'dy);
  e0=zeros(n,(k+1));
  e1=zeros(n,(k+1));
  beta0=zeros(k+2+p,k+1);
  beta1=zeros(k+2+p,k+1);
  for j (0,k,1);
   r0=zeros(k+2+p,2);
   r0[p+1,1]=1;
   r0[p+2,2]=2;
   if j<k;
     r1=zeros(2+p+j,k-j)|eye(k-j);
     b1=beta-xxi*r1*invpd(r1'xxi*r1)*(r1'beta);
     b1[2+p+j+1:2+p+k]=zeros(k-j,1);
     r00=r0~r1;
     b0=beta-xxi*r00*invpd(r00'xxi*r00)*(r00'beta);
     b0[p+1:p+2]=0|0;
     b0[2+p+j+1:2+p+k]=zeros(k-j,1);
   else;
     b1=beta;
     b0=beta-xxi*r0*invpd(r0'xxi*r0)*(r0'beta);
     b0[p+1:p+2]=0|0;
   endif;
   e0[.,j+1]=dy-x*b0;
   e1[.,j+1]=dy-x*b1;
   beta0[.,j+1]=b0;
   beta1[.,j+1]=b1;
  endfor;
  aa=moment(e0~e1,0);
  kk=2*(k+1);
  sig=aa[kk,kk]/(n-k-2+p);
  r=(seqa(p,1,k+1)|seqa(2+p,1,k+1))*sig;
  alpha0=ones(kk,1)/kk;
  {w,u1,u2,u3,u4,ret}=QProg(alpha0,aa,-r,ones(1,kk),1,0,0,0~1); 
  w=w.*(w .> 1e-20);
  w=w./sumc(w);
  beta=(beta0~beta1)*w;
retp(beta,w);
endp;
